import { __awaiter } from 'tslib';

const Symbol_observable = (() => (typeof Symbol === 'function' && Symbol.observable) || '@@observable')();

const noop = () => { };
const identity = (x) => x;
function doCleanup(cleanup) {
    if (typeof cleanup === 'function') {
        cleanup();
    }
    else if (typeof cleanup === 'object') {
        if ('cancel' in cleanup) {
            cleanup.cancel();
        }
        else if ('unsubscribe' in cleanup) {
            cleanup.unsubscribe();
        }
        else if ('dispose' in cleanup) {
            cleanup.dispose();
        }
    }
}
/* istanbul ignore next */
function panic(code) {
    throw new Error(`Fatal error code (${code}): please file an issue to report this.`);
}

class Suspend {
    constructor(cancel) {
        this.cancel = cancel;
    }
}
class SuspendWithFallback extends Suspend {
    constructor(fallback, cancel) {
        super(cancel);
        this.fallback = fallback;
        this.cancel = cancel;
    }
}
let currentCollecting = null;
function setData(data, value, equalCheck) {
    if (currentCollecting) {
        console.error(`Violated action: You can't mutate any behavior inside a computation.`);
        return;
    }
    if (!inTransaction) {
        return transaction(() => setData(data, value, equalCheck));
    }
    if (equalCheck && data.value === value) {
        return;
    }
    if (data.flags & 8 /* Changed */) {
        console.error(`You can't set a Cell twice in a transaction.`);
        return;
    }
    data.value = value;
    data.flags |= 8 /* Changed */;
    addDirtyData(data);
}
const effects = [];
let last = null;
function addDirtyData(data) {
    if (last === null) {
        last = {
            prev_observer: null,
            next_observer: null,
            observer: data,
        };
    }
    else {
        const tmp = last;
        last.next_observer = last = {
            prev_observer: tmp,
            next_observer: null,
            observer: data,
        };
    }
}
function markObserversForCheck(data) {
    const stack = [data];
    while (stack.length > 0) {
        const node = stack.pop();
        let p = node.last_observer;
        while (p !== null) {
            const observer = p.observer;
            p = p.prev_observer; // go_next
            observer.depsCounter++;
            if (observer.flags & 32 /* MarkForCheck */) {
                continue;
            }
            observer.flags |= 32 /* MarkForCheck */;
            stack.push(observer);
        }
    }
}
function accessData(data) {
    if (currentCollecting !== null) {
        logRead(currentCollecting, data);
    }
    return data.value;
}
function accessComputation(data) {
    if (data.flags & 256 /* Lazy */) {
        panic(4);
    }
    if (data.flags & 16 /* Estimating */) {
        // self referenced estimation.
        return data.value; // prev_state
    }
    if (currentCollecting !== null) {
        logRead(currentCollecting, data);
        if (currentCollecting.flags & 64 /* Suspensed */ &&
            data.flags & 128 /* Suspending */) {
            throw new Suspend(noop); //TODO: current cancel?
        }
    }
    if (inTransaction) {
        data.flags |= 16 /* Estimating */;
        const value = data.collect();
        data.flags &= ~16 /* Estimating */;
        return value;
    }
    if (data.flags & 4 /* Stale */) {
        estimateComputation(data, data.collect);
    }
    return data.value; // current value
}
function estimateComputation(comp, expr) {
    if ((comp.flags & 4 /* Stale */) === 0) {
        panic(10);
    }
    comp.flags |= 16 /* Estimating */;
    comp.flags &= ~4 /* Stale */;
    const value = computeAndCollect(comp, expr);
    comp.value = value;
    comp.flags &= ~16 /* Estimating */;
}
function logRead(accessor, data) {
    if (accessor.checkNode === null) {
        insertNewSource(accessor, data);
    }
    else if (accessor.checkNode.source !== data) {
        // TODO: refactor
        const checkEnd = accessor.last_source;
        accessor.last_source = accessor.checkNode.prev_source;
        accessor.last_source.next_source = null; // new last_source
        const checkStart = accessor.checkNode;
        checkStart.prev_source = null;
        accessor.cleanNode = checkEnd;
        accessor.checkNode = null;
        insertNewSource(accessor, data);
    }
    else {
        accessor.checkNode = accessor.checkNode.next_source;
    }
}
function untrack(fn, ...args) {
    const stored = currentCollecting;
    currentCollecting = null;
    const ret = fn(...args);
    currentCollecting = stored;
    return ret;
}
function computeAndCollect(computation, computeExpr) {
    var _a, _b;
    const stored = currentCollecting;
    currentCollecting = computation;
    computation.checkNode = computation.first_source;
    let currentValue;
    if ((computation.flags & 64 /* Suspensed */) === 0) {
        // [likely]
        try {
            currentValue = computeExpr(computation.value);
        }
        catch (e) {
            const d = new Error('Error caught in computed(): ' + e);
            if (e.stack)
                d.stack = e.stack;
            throw d;
        }
    }
    else {
        (_b = (_a = computation).currentCancel) === null || _b === void 0 ? void 0 : _b.call(_a); // if is suspending, cancel last (maybe no effect)
        computation.currentCancel = undefined;
        try {
            computation.flags |= 128 /* Suspending */;
            currentValue = computeExpr(computation.value);
            computation.flags &= ~128 /* Suspending */;
        }
        catch (e) {
            if (e instanceof Suspend) {
                // a _behavior_ should have been collected and it will schedule an update later.
                if (e instanceof SuspendWithFallback) {
                    currentValue = e.fallback;
                }
                else {
                    currentValue = computation.fallback;
                }
                computation.currentCancel = e.cancel;
            }
            else {
                const d = new Error('Error caught in suspended(): ' + e);
                if (e.stack)
                    d.stack = e.stack;
                throw d;
            }
        }
        finally {
        }
    }
    if (computation.checkNode !== null) {
        // collected source num is less than expected
        // but it's fine as we remove the extra sources.
        cleanupSources(computation.last_source, computation.checkNode.prev_source);
        computation.last_source = computation.checkNode.prev_source;
        if (computation.last_source) {
            computation.last_source.next_source = null;
        }
        else {
            computation.first_source = null;
            // has no source.
        }
        computation.checkNode = null;
    }
    if (computation.cleanNode) {
        // these nodes are detached from observer.
        cleanupSources(computation.cleanNode, null);
        computation.cleanNode = null;
    }
    currentCollecting = stored;
    return currentValue;
}
function cleanupComputation(computation) {
    cleanupSources(computation.last_source, null);
    computation.first_source = null;
    computation.last_source = null;
    computation.flags |= 4 /* Stale */;
}
function cleanupSources(from, until) {
    let lastSource = from;
    while (lastSource !== until) {
        const obRef = lastSource.observer_ref;
        if (obRef.next_observer) {
            obRef.next_observer.prev_observer = obRef.prev_observer;
        }
        else {
            lastSource.source.last_observer = obRef.prev_observer;
        }
        if (obRef.prev_observer) {
            obRef.prev_observer.next_observer = obRef.next_observer;
        }
        obRef.observer = null;
        // last source
        const source = lastSource.source;
        if (source.flags & 2 /* Computation */ &&
            source.last_observer == null &&
            source.last_effect == null) {
            cleanupComputation(source);
        }
        lastSource = lastSource.prev_source;
        if (lastSource) {
            lastSource.next_source = null;
        }
    }
}
function propagate(origin) {
    const stack = [origin];
    while (stack.length > 0) {
        const data = stack[stack.length - 1];
        if (data.flags & 512 /* Propagating */) {
            stack.pop();
            data.flags -= 512 /* Propagating */;
            // do cleanup
            if (data.flags & 2 /* Computation */) {
                if (data.last_observer === null /* unlikely */ &&
                    data.last_effect === null) {
                    cleanupComputation(data);
                }
            }
            continue;
        }
        data.flags |= 512 /* Propagating */;
        let observer = data.last_observer;
        const changes = (data.flags & 8 /* Changed */) >> 1; // 8>>1 = 4
        while (observer !== null) {
            const current = observer.observer;
            observer = observer.prev_observer;
            current.flags |= changes; // bug: data can be stale, too.
            current.depsCounter--;
            if (current.depsCounter === 0) {
                stack.push(current);
                current.flags &= ~32 /* MarkForCheck */;
                if (current.flags & 4 /* Stale */) {
                    current.flags &= ~4 /* Stale */;
                    if (current.flags & 256 /* Lazy */) {
                        // lazy comp will keep stale? until re-estimated.
                        current.flags |= 8 /* Changed */ | 4 /* Stale */;
                    }
                    else if (current.flags & 2 /* Computation */) {
                        const currentValue = computeAndCollect(current, current.collect);
                        // compare value , if changed, mark as changed
                        if (currentValue !== current.value) {
                            current.value = currentValue;
                            current.flags |= 8 /* Changed */;
                        }
                    }
                }
            }
        }
        if (changes) {
            data.flags &= ~8 /* Changed */;
            let watcher = data.last_effect;
            while (watcher !== null) {
                effects.push(watcher.fn);
                watcher = watcher.prev;
            }
        }
    }
}
function watch$1(data, sideEffect) {
    if (currentCollecting) {
        console.error(`You should not watch inside computation.`);
        return { disposed: true };
    }
    const node = {
        fn: sideEffect,
        prev: data.last_effect,
        next: null,
        disposed: false,
        data: data,
    };
    if (data.last_effect) {
        data.last_effect.next = node;
    }
    data.last_effect = node;
    if (data.flags & 4 /* Stale */ && (data.flags & 256 /* Lazy */) === 0) {
        // must do estimation to get deps change propagated.
        estimateComputation(data, data.collect);
    }
    return node;
}
function disposeWatcher(watcher) {
    if (watcher.disposed) {
        return;
    }
    watcher.disposed = true;
    if (watcher.next === null) {
        // it is the last.
        watcher.data.last_effect = watcher.prev;
    }
    else {
        watcher.next.prev = watcher.prev;
    }
    if (watcher.prev) {
        watcher.prev.next = watcher.next;
    }
}
let inTransaction = false;
function transaction(fn) {
    if (inTransaction) {
        // already inside a transaction
        return fn();
    }
    inTransaction = true;
    const retValue = fn();
    inTransaction = false;
    const data = {
        flags: 8 /* Changed */ | 32 /* MarkForCheck */,
        last_observer: last,
        last_effect: null,
        value: undefined,
        depsCounter: 0,
    };
    last = null;
    markObserversForCheck(data);
    propagate(data);
    while (effects.length > 0) {
        effects.pop()();
    }
    return retValue;
}
function insertNewSource(accessing, source) {
    if (accessing.last_source !== null) {
        const node = {
            source: source,
            prev_source: accessing.last_source,
            next_source: null,
            observer_ref: insertNewObserver(source, accessing),
        };
        accessing.last_source.next_source = node;
        accessing.last_source = node;
    }
    else {
        const node = {
            source: source,
            prev_source: null,
            next_source: null,
            observer_ref: insertNewObserver(source, accessing),
        };
        accessing.first_source = node; // if last is null, first is definitely null.
        accessing.last_source = node;
    }
}
function insertNewObserver(accesed, observer) {
    if (accesed.last_observer !== null) {
        const node = {
            observer: observer,
            prev_observer: accesed.last_observer,
            next_observer: null,
        };
        accesed.last_observer.next_observer = node;
        accesed.last_observer = node;
        return node;
    }
    else {
        const node = {
            observer: observer,
            prev_observer: null,
            next_observer: null,
        };
        accesed.last_observer = node;
        return node;
    }
}
function createLazy(initial) {
    const ret = {
        flags: 2 /* Computation */ | 256 /* Lazy */ | 4 /* Stale */,
        last_effect: null,
        last_observer: null,
        value: undefined,
        checkNode: null,
        cleanNode: null,
        first_source: null,
        last_source: null,
        collect: null,
        depsCounter: 0,
    };
    if (initial !== undefined) {
        ret.value = initial;
    }
    return ret;
}
function executeLazy(computation, fn) {
    if (computation.flags & 4 /* Stale */) {
        estimateComputation(computation, fn);
    }
    return fn(computation.value); // TODO: type not correct
}
function createData(value) {
    return {
        flags: 1 /* Data */,
        last_effect: null,
        last_observer: null,
        depsCounter: 0,
        value,
    };
}
function createComputation(fn, options) {
    const ret = {
        flags: 2 /* Computation */ | 4 /* Stale */,
        last_effect: null,
        last_observer: null,
        depsCounter: 0,
        value: undefined,
        checkNode: null,
        cleanNode: null,
        first_source: null,
        last_source: null,
        collect: fn,
    };
    if (options) {
        if (options.initial !== undefined) {
            ret.value = options.initial;
        }
    }
    return ret;
}
function createSuspended(fn, fallback) {
    const ret = {
        flags: 2 /* Computation */ | 64 /* Suspensed */ | 4 /* Stale */,
        last_effect: null,
        last_observer: null,
        depsCounter: 0,
        value: undefined,
        checkNode: null,
        cleanNode: null,
        first_source: null,
        last_source: null,
        collect: fn,
        fallback,
        currentCancel: undefined,
    };
    return ret;
}
function __current_collecting() {
    return currentCollecting;
}

class Cell {
    constructor(internal) {
        this.internal = internal;
    }
    get value() {
        return accessData(this.internal);
    }
    [Symbol_observable]() {
        return this;
    }
    *[Symbol.iterator]() {
        return yield (resolve) => {
            const unsub = this.watch((value) => {
                resolve(value);
                unsub();
            });
            return unsub;
        };
    }
    map(mapFn) {
        const internal = createComputation(() => mapFn(this.value));
        return new ComputationalCell(internal);
    }
    watch(watchFn, options) {
        let lastDisposer = undefined;
        const watcher = watch$1(this.internal, () => {
            doCleanup(lastDisposer);
            lastDisposer = watchFn(this.internal.value);
        });
        if (options === null || options === void 0 ? void 0 : options.immediate) {
            lastDisposer = watchFn(this.internal.value);
        }
        return () => {
            doCleanup(lastDisposer);
            disposeWatcher(watcher);
        };
    }
    /**
     * @deprecated Use watch.
     */
    subscribe(next) {
        const ret = this.watch((v) => {
            next(v);
        }, { immediate: true });
        ret.unsubscribe = ret;
        return ret;
    }
}
class ComputationalCell extends Cell {
    constructor(internal) {
        super(internal);
    }
    get value() {
        return accessComputation(this.internal);
    }
}
class Lazy {
    constructor(internal) {
        this.internal = internal;
    }
    get stale() {
        return (this.internal.flags & 4 /* Stale */) > 0;
    }
    execute(expr) {
        return executeLazy(this.internal, expr);
    }
    watch(watchFn, options) {
        let lastDisposer = undefined;
        const watcher = watch$1(this.internal, () => {
            doCleanup(lastDisposer);
            lastDisposer = watchFn(this.internal.value);
        });
        if (options === null || options === void 0 ? void 0 : options.immediate) {
            lastDisposer = watchFn(this.internal.value);
        }
        return () => {
            doCleanup(lastDisposer);
            disposeWatcher(watcher);
        };
    }
}
function mutable(initialValue) {
    const internal = createData(initialValue);
    return [
        new Cell(internal),
        (v) => {
            if (v instanceof Function) {
                setData(internal, v(internal.value), true);
            }
            else {
                setData(internal, v, true);
            }
        },
    ];
}
function mutValue(initialValue) {
    const internal = createData(initialValue);
    return [
        new Cell(internal),
        (v) => {
            setData(internal, v, true);
        },
    ];
}
function lazy(initial) {
    return new Lazy(createLazy(initial));
}
function computed(expr, options) {
    const internal = createComputation(expr, options);
    return new ComputationalCell(internal);
}
function suspended(expr, fallback) {
    const internal = createSuspended(expr, fallback);
    return new ComputationalCell(internal);
}
function constant(value) {
    return new Cell(createData(value));
}
function combined(obj, lifeFn = identity) {
    if (obj instanceof Array) {
        return computed(() => {
            return lifeFn(obj.map((x) => {
                return x.value;
            }));
        });
    }
    return computed(() => {
        return lifeFn(Object.fromEntries(Object.entries(obj).map(([key, value]) => {
            return [key, value.value];
        })));
    });
}

function mutArray(initial) {
    const array = initial;
    const internal = createData(array);
    return [
        new Cell(internal),
        {
            set length(value) {
                array.length = value;
                setData(internal, array, false);
                return;
            },
            get length() {
                return array.length;
            },
            setAt(index, value) {
                array[index] = value;
                setData(internal, array, false);
                return;
            },
            push(...items) {
                const ret = array.push(...items);
                setData(internal, array, false);
                return ret;
            },
            pop() {
                const ret = array.pop();
                setData(internal, array, false);
                return ret;
            },
            shift() {
                const ret = array.shift();
                setData(internal, array, false);
                return ret;
            },
            unshift(...items) {
                const ret = array.unshift(...items);
                setData(internal, array, false);
                return ret;
            },
            splice(start, deleteCount) {
                const ret = array.splice(start, deleteCount);
                setData(internal, array, false);
                return ret;
            },
            reverse() {
                const ret = array.reverse();
                setData(internal, array, false);
                return ret;
            },
        },
    ];
}
function mutSet(initial) {
    const set = new Set(initial);
    const internal = createData(set);
    return [
        new Cell(internal),
        {
            add(value) {
                set.add(value);
                setData(internal, set, false);
                return this;
            },
            delete(value) {
                const ret = set.delete(value);
                setData(internal, set, false);
                return ret;
            },
            clear() {
                set.clear();
                setData(internal, set, false);
            },
        },
    ];
}
function mutMap(initial) {
    const map = new Map(initial ? [...initial] : []);
    const internal = createData(map);
    return [
        new Cell(internal),
        {
            clear() {
                map.clear();
                setData(internal, map, false);
            },
            delete(key) {
                const ret = map.delete(key);
                setData(internal, map, false);
                return ret;
            },
            set(key, value) {
                map.set(key, value);
                setData(internal, map, false);
                return this;
            },
        },
    ];
}

const END = {};
function eventReaderBase(from, until) {
    let closed = false;
    const bufferQueue = [];
    let continuation = null;
    const stopUntilListener = until.listen((next) => {
        if (!closed) {
            dispose();
        }
    });
    const stopFromListener = from.listen((next) => {
        if (continuation) {
            continuation(next);
        }
        else {
            bufferQueue.push(next);
        }
    });
    function dispose() {
        if (!closed) {
            stopUntilListener();
            stopFromListener();
            if (continuation) {
                continuation(END);
                continuation = null;
            }
            closed = true;
        }
    }
    const next = function* () {
        /* istanbul ignore if */
        if (continuation !== null) {
            throw new Error(`There exists a Task waiting for this channel already.`);
        }
        if (closed) {
            throw new Error(`Closed`);
        }
        if (bufferQueue.length > 0) {
            return bufferQueue.shift();
        }
        return (yield (resolve, reject) => {
            continuation = (value) => {
                if (value === END) {
                    reject(new Error(`Closed`));
                    return;
                }
                continuation = null;
                resolve(value);
            };
            return () => {
                continuation = null;
            };
        });
    };
    const hasNext = function* () {
        /* istanbul ignore if */
        if (continuation !== null) {
            throw new Error(`There exists a Task waiting for this channel already.`);
        }
        if (closed) {
            return false;
        }
        if (bufferQueue.length > 0) {
            return true;
        }
        return yield (resolve) => {
            continuation = (value) => {
                if (value === END) {
                    continuation = null;
                    resolve(false);
                    return;
                }
                bufferQueue.push(value);
                continuation = null;
                resolve(true);
            };
            return () => {
                continuation = null;
            };
        };
    };
    return {
        next,
        hasNext,
        dispose,
    };
}
function readEvents(props) {
    return eventReaderBase(props.from, props.until);
}

class CanceledError extends Error {
    constructor(message) {
        super(message !== null && message !== void 0 ? message : 'Operation has been canceled');
        this.name = 'CanceledError';
    }
}
let $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER = null;
class CancellablePromise extends Promise {
    constructor(executor) {
        super((resolve, reject) => {
            let settled = false;
            // disposer is guaranteed to be not executed after: 1.fulfilled 2.rejected 3.disposed once(which is also reject)
            // resolve in dispose? I think it's valid. And default behavior is reject after disposer executed, but user
            // can reject/resolve in disposer and default behavior is ignored.
            const dispose = executor((v) => {
                resolve(v);
                settled = true;
            }, (e) => {
                reject(e);
                settled = true;
            });
            $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER = () => {
                if (settled) {
                    return;
                }
                if (dispose) {
                    doCleanup(dispose); // might become settled.
                    settled = true;
                    reject(new CanceledError());
                }
            };
        });
        this.cancel = $$CANT_ACCESS_CHILD_PROP_INSIDE_SUPER;
    }
    *[Symbol.iterator]() {
        return (yield this);
    }
}

function executeRunnable(runnable, onsuccess, onfailure) {
    let currentDisposer = null;
    const iter = runnable[Symbol.iterator]();
    let settled = false;
    function takeControl(yieldedObject) {
        try {
            if (yieldedObject instanceof CancellablePromise) {
                currentDisposer = () => yieldedObject.cancel();
                yieldedObject.then(resumeTask, throwTask);
            }
            else if (typeof yieldedObject === 'function') {
                //bind sideEffect
                let synchronous = true; // closure callback
                let innerSettled = false;
                let synchronousResult = null;
                function tryResolve(value) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$RESOLVE: true,
                            $$VALUE: value,
                        };
                    }
                    else {
                        resumeTask(value);
                    }
                }
                function tryReject(error) {
                    if (innerSettled)
                        return;
                    innerSettled = true;
                    if (synchronous) {
                        synchronousResult = {
                            $$REJECT: true,
                            $$VALUE: error,
                        };
                    }
                    else {
                        throwTask(error);
                    }
                }
                const maybeDisposer = yieldedObject(tryResolve, tryReject); //might raise an error.
                if (synchronousResult !== null) {
                    throw synchronousResult;
                }
                currentDisposer = () => {
                    // if (innerSettled) return; // no need this line: currentDisposer
                    // is guaranteed to execute only once.
                    doCleanup(maybeDisposer); // maybe raise an error.
                    innerSettled || tryReject(new CanceledError());
                    // innerSettled = true;
                };
                synchronous = false;
            }
            else {
                throw new TypeError('Invalid object yielded. Are you missing an asterisk(*) after `yield`?');
            }
        }
        catch (e) {
            if (e.$$RESOLVE === true) {
                return resumeTask(e.$$VALUE);
            }
            else if (e.$$REJECT === true) {
                return throwTask(e.$$VALUE);
            }
            else {
                return throwTask(e);
            }
        }
    }
    function resumeTask(resumeValue) {
        currentDisposer = null;
        let result;
        try {
            result = iter.next(resumeValue);
        }
        catch (e) {
            return throwTask(e);
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        else {
            takeControl(result.value);
        }
    }
    function throwTask(error) {
        currentDisposer = null;
        let result;
        try {
            result = iter.throw(error);
        }
        catch (e) {
            if (error === e) {
                settled || ((settled = true), onfailure(e));
                return;
            }
            else {
                return throwTask(e);
            }
        }
        if (result.done) {
            settled || ((settled = true), onsuccess(result.value));
            return;
        }
        takeControl(result.value);
    }
    resumeTask(undefined); //start synchronously.
    return () => {
        if (!settled) {
            try {
                currentDisposer === null || currentDisposer === void 0 ? void 0 : currentDisposer();
            }
            catch (e) {
                if (!(e instanceof CanceledError)) {
                    throw e;
                }
            }
            finally {
                settled = true;
            }
        }
    };
}
// bug: type inference breaks when calling itself.
function task(taskFn) {
    return function (...params) {
        return new CancellablePromise((resolve, reject) => {
            return executeRunnable(taskFn.call(undefined, ...params), resolve, reject);
        });
    };
}
function start(runnable) {
    return new CancellablePromise((resolve, reject) => {
        return executeRunnable(runnable, resolve, reject);
    });
}
function* delay(time) {
    return (yield (resume) => {
        const id = setTimeout((s) => resume(s), time);
        return () => clearTimeout(id);
    });
}
function* timeout(time) {
    return (yield (_, reject) => {
        const id = setTimeout((s) => reject(new Error('Timeout')), time);
        return () => clearTimeout(id);
    });
}
/* istanbul ignore next */
function* nextAnimationFrame() {
    return (yield (resume) => {
        const id = requestAnimationFrame(resume);
        return () => cancelAnimationFrame(id);
    });
}
class Semaphore {
    constructor(maxConcurrency) {
        this.maxConcurrency = maxConcurrency;
        this.currentNum = 0;
        this.queue = [];
    }
    get free() {
        return this.currentNum < this.maxConcurrency;
    }
    *waitOne() {
        if (this.currentNum >= this.maxConcurrency) {
            yield* new CancellablePromise((resolve) => {
                this.queue.push(resolve);
                return () => this.queue.splice(this.queue.indexOf(resolve), 1); // resolve definitely exist.
            });
            // a free space is gurantted.
        }
        this.currentNum++;
        return;
    }
    // this is an action
    release() {
        if (this.currentNum === 0) {
            return;
        }
        this.currentNum--;
        // trigger?
        if (this.queue.length) {
            this.queue.shift()(); // a microtask is scheduled.
        }
    }
}
function* resolve(obj) {
    if (typeof obj === 'object' && obj !== null) {
        if (Symbol.iterator in obj) {
            return yield* obj;
        }
        else if ('then' in obj) {
            return yield (resolve, reject) => {
                obj.then(resolve, reject);
                return noop;
            };
        }
        else if (Symbol_observable in obj) {
            return yield (resolve, reject) => {
                let lastEmit = undefined;
                const subscription = obj[Symbol_observable]().subscribe((next) => {
                    lastEmit = next;
                }, (error) => reject(error), () => {
                    resolve(lastEmit);
                });
                return () => subscription.unsubscribe();
            };
        }
        else {
            return yield (_, reject) => {
                reject(new TypeError("Can't resolve object"));
                return noop;
            };
        }
    }
    else {
        return yield (resolve) => {
            resolve(obj);
            return noop;
        };
    }
}
function* any(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const errors = [];
        const disposers = [];
        let sync = true;
        let errorCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    if (sync)
                        throw value;
                    success(value);
                    disposers.forEach((x) => x()); // async
                }, (error) => {
                    errorCount++;
                    errors[currentIndex] = error;
                    if (!sync) {
                        if (errorCount === count) {
                            fail(new AggregateError(errors, 'All task failed'));
                        }
                    }
                }));
            }
            catch (value) {
                success(value);
                disposers.forEach((x) => x()); // sync
                return;
            }
            count++;
        }
        sync = false;
        if (errorCount === count) {
            return fail(new AggregateError(errors, 'All task failed'));
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
function* race(tasks) {
    return yield (success, fail) => {
        const disposers = [];
        let sync = true;
        for (const task of tasks) {
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    if (sync) {
                        throw {
                            $$RESOLVE: true,
                            $$VALUE: value,
                        };
                    }
                    success(value);
                    disposers.forEach((x) => x()); // async
                }, (error) => {
                    if (sync)
                        throw {
                            $$REJECT: true,
                            $$VALUE: error,
                        };
                    fail(error);
                    disposers.forEach((x) => x()); // async
                }));
            }
            catch (value) {
                if (value.$$RESOLVE === true) {
                    success(value.$$VALUE);
                }
                else if (value.$$REJECT === true) {
                    fail(value.$$VALUE);
                }
                else {
                    fail(value); // highly impossible, but for insurance
                }
                sync = false; // NB
                disposers.forEach((x) => x()); // sync
                return;
            }
        }
        sync = false;
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
// it's the opposite of any!
function* all(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const result = [];
        const disposers = [];
        let sync = true;
        let succeedCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    succeedCount++;
                    result[currentIndex] = value;
                    if (!sync) {
                        if (succeedCount === count) {
                            success(result);
                        }
                    }
                }, (error) => {
                    if (sync)
                        throw error;
                    fail(error);
                    disposers.forEach((x) => x()); //async
                }));
            }
            catch (e) {
                fail(e);
                sync = false;
                disposers.forEach((x) => x()); //sync
                return;
            }
            count++;
        }
        sync = false;
        if (succeedCount === count) {
            return success(result);
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}
function* allSettled(tasks) {
    return yield (success, fail) => {
        let count = 0;
        const result = [];
        const disposers = [];
        let sync = true;
        let settledCount = 0;
        for (const task of tasks) {
            const currentIndex = count;
            try {
                disposers.push(executeRunnable(resolve(task), (value) => {
                    settledCount++;
                    result[currentIndex] = {
                        success: true,
                        value,
                    };
                    if (!sync) {
                        if (settledCount === count) {
                            success(result);
                        }
                    }
                }, (error) => {
                    settledCount++;
                    result[currentIndex] = {
                        success: false,
                        value: error,
                    };
                    if (!sync) {
                        if (settledCount === count) {
                            success(result);
                        }
                    }
                }));
            }
            catch (e) {
                // almost you can catch nothing, but for insurance
                fail(e);
                disposers.forEach((x) => x());
                return;
            }
            count++;
        }
        sync = false;
        if (settledCount === count) {
            return success(result);
        }
        return () => {
            fail(new CanceledError());
            disposers.forEach((x) => x());
        };
    };
}

const BITS_PER_BYTE = 32;
const MASK = BITS_PER_BYTE - 1; //
class BloomFilter {
    constructor(size, numOfHash, inheritFrom) {
        this.size = size;
        this.numOfHash = numOfHash;
        if (size <= 0) {
            throw Error('size should be positive');
        }
        // size should be the power of 2
        if (size & (size - 1)) {
            throw Error('size should be power of 2');
        }
        if (typeof Uint32Array !== undefined) {
            this.buckets = new Uint32Array(size);
        }
        else {
            this.buckets = new Array(size).fill(0); // as if it is
        }
        this.m = this.size * BITS_PER_BYTE;
        if (inheritFrom) {
            for (let i = 0; i < size; i++) {
                this.buckets[i] = inheritFrom[i];
            }
        }
    }
    add(item) {
        const locations = this.locations(item);
        for (let i = 0; i < this.numOfHash; i++) {
            const num = locations[i];
            this.buckets[num >> 5] |= 1 << (num & MASK);
        }
    }
    test(item) {
        const locations = this.locations(item);
        for (let i = 0; i < this.numOfHash; i++) {
            const num = locations[i];
            if ((this.buckets[num >> 5] & (1 << (num & MASK))) === 0) {
                return false;
            }
        }
        return true;
    }
    locations(item) {
        const a = fnv_1a(item);
        const b = fnv_1a(item, 1576284489); // The seed value is chosen randomly
        const m = this.m;
        let x = a & (m - 1);
        const r = [];
        for (let i = 0; i < this.numOfHash; i++) {
            r.push(x < 0 ? x + m : x);
            x = (x + b) & (m - 1);
        }
        return r;
    }
}
// Steal from https://github.com/jasondavies/bloomfilter.js/blob/master/bloomfilter.js
// Fowler/Noll/Vo hashing.
// Nonstandard variation: this function optionally takes a seed value that is incorporated
// into the offset basis. According to http://www.isthe.com/chongo/tech/comp/fnv/index.html
// "almost any offset_basis will serve so long as it is non-zero".
function fnv_1a(v, seed = 0) {
    let a = 2166136261 ^ seed;
    for (let i = 0, n = v.length; i < n; ++i) {
        const c = v.charCodeAt(i), d = c & 0xff00;
        if (d)
            a = fnv_multiply(a ^ (d >> 8));
        a = fnv_multiply(a ^ (c & 0xff));
    }
    return fnv_mix(a);
}
// a * 16777619 mod 2**32
function fnv_multiply(a) {
    return a + (a << 1) + (a << 4) + (a << 7) + (a << 8) + (a << 24);
}
// See https://web.archive.org/web/20131019013225/http://home.comcast.net/~bretm/hash/6.html
function fnv_mix(a) {
    a += a << 13;
    a ^= a >>> 7;
    a += a << 3;
    a ^= a >>> 17;
    a += a << 5;
    return a & 0xffffffff;
}

/**
 * The dirtiest part of kairo (the second is concurrency...)
 * The bloomfilter process is inspired by Angular.
 */
class Scope {
    constructor(parentScope, rootScope) {
        this.parentScope = parentScope;
        this.rootScope = rootScope;
        this.onmountLogics = [];
        this.injections = new Map();
        this.attached = false;
        this.injections_bloom = new BloomFilter(8, 4, parentScope === null || parentScope === void 0 ? void 0 : parentScope.injections_bloom.buckets);
        this.rootScope = rootScope !== null && rootScope !== void 0 ? rootScope : parentScope === null || parentScope === void 0 ? void 0 : parentScope.rootScope;
    }
    beginScope() {
        const stored = Scope._currentScope;
        Scope._currentScope = this;
        return () => {
            Scope._currentScope = stored;
        };
    }
    static get current() {
        if (Scope._currentScope) {
            return Scope._currentScope;
        }
        /* istanbul ignore next */
        throw new TypeError('Not inside a scope.');
    }
    registerOnMountLogic(logic) {
        /* istanbul ignore if */
        if (this !== Scope.current) {
            throw new TypeError('Invalid opearation');
        }
        this.onmountLogics.push(logic);
    }
    registerProvider(arg0, arg1, ...args) {
        /* istanbul ignore if */
        if (this !== Scope.current) {
            throw new TypeError('Invalid opearation');
        }
        if (arg0 instanceof Token) {
            this.injections_bloom.add(arg0.name);
            this.injections.set(arg0, arg1);
            return arg1;
        }
        else if (typeof arg0 === 'function' && assertType()) {
            const exposed = arg0(arg1, ...args);
            this.injections_bloom.add(arg0.name);
            this.injections.set(arg0, exposed);
            return exposed;
        }
    }
    inject(token, options) {
        var _a, _b, _c;
        let scope = this;
        if (options === null || options === void 0 ? void 0 : options.skipSelf) {
            scope = (_a = scope.parentScope) !== null && _a !== void 0 ? _a : scope.rootScope; // in case current scope is top
        }
        // get bloom hash
        if ((scope === null || scope === void 0 ? void 0 : scope.injections_bloom.test(token.name)) === false) {
            // if not exist in bloom filter, directly find in root
            scope = scope.rootScope;
        }
        while (scope) {
            const d = scope.injections.get(token);
            if (d === undefined) {
                scope = (_b = scope.parentScope) !== null && _b !== void 0 ? _b : scope.rootScope;
            }
            else {
                return d;
            }
        }
        if (options === null || options === void 0 ? void 0 : options.optional) {
            return (_c = options.defaultValue) !== null && _c !== void 0 ? _c : undefined;
        }
        throw Error(`Injection token '${token.name}' is not found.`);
    }
    attach() {
        /* istanbul ignore if */
        if (this.attached) {
            throw TypeError('Scope has been attached');
        }
        this.attached = true;
        const cleanups = this.onmountLogics.map((x) => {
            const cleanup = x();
            if (cleanup instanceof CancellablePromise) {
                cleanup.catch(noop); // to avoid unhandled promise rejection
            }
            return cleanup;
        });
        return () => {
            this.attached = false;
            cleanups.forEach(doCleanup);
        };
    }
}
Scope._currentScope = null;
function mount(onmount) {
    Scope.current.registerOnMountLogic(onmount);
}
class Token {
    constructor(name) {
        this.name = name;
    }
    static for(name) {
        return new Token(name);
    }
    /* istanbul ignore next */
    toString() {
        return this.name;
    }
}
function assertType(value) {
    return true;
}
function provide(arg0, ...arg1) {
    return Scope.current.registerProvider(arg0, ...arg1);
}
function inject(token, options) {
    return Scope.current.inject(token, options);
}

class EventStream {
    constructor(producer) {
        this.producer = producer;
        this.disposeProducer = null;
        this.head = null;
        this.tail = null;
        this.propagating = false;
    }
    *[Symbol.iterator]() {
        return yield (resolve) => {
            const unsub = this.listen((value) => {
                resolve(value);
                unsub(); // 
            });
            return unsub;
        };
    }
    __internal_subscribe(handler) {
        const node = {
            next: null,
            prev: this.tail,
            handler: handler,
        };
        if (!this.head) {
            this.head = node;
            this.tail = node;
            // activate
            this.disposeProducer = this.producer((s) => this.next(s));
        }
        else {
            this.tail.next = node;
            this.tail = node;
        }
        const subscription = () => {
            if (node.prev) {
                //node is head
                node.prev.next = node.next;
            }
            else {
                this.head = node.next;
            }
            if (node.next) {
                //node is tail
                node.next.prev = node.prev;
            }
            else {
                this.tail = node.prev;
            }
            if (!this.head) {
                // deactivate
                doCleanup(this.disposeProducer);
            }
        };
        return subscription;
    }
    next(value) {
        if (this.propagating) {
            console.error(`A loop occured.`); // TODO: detailed error message.
            return;
        }
        this.propagating = true;
        let p = this.head;
        const handlers = [];
        while (p !== null) {
            handlers.push(p.handler);
            p = p.next;
        }
        // linked list is mutable data structure thus a immutable snapshot of subscriptions is required.
        // then you can subscribe or unsubscirbe inside a handler.
        while (handlers.length) {
            handlers.pop()(value);
        }
        this.propagating = false;
    }
    listen(handler) {
        let lastDisposer = undefined;
        const unsubscribe = this.__internal_subscribe((value) => {
            doCleanup(lastDisposer);
            lastDisposer = handler(value);
        });
        return () => {
            doCleanup(lastDisposer);
            unsubscribe();
        };
    }
    transform(transformFn) {
        return new EventStream((next) => {
            return this.__internal_subscribe((v) => next(transformFn(v)));
        });
    }
    filter(filterFn) {
        return new EventStream((next) => {
            return this.__internal_subscribe((v) => { var _a; return (_a = filterFn(v)) !== null && _a !== void 0 ? _a : next(v); });
        });
    }
    static create() {
        const stream = new EventStream(() => {
            return () => { };
        });
        return [stream, (payload) => transaction(() => stream.next(payload))];
    }
}
function stream() {
    return EventStream.create();
}
function never() {
    const [stream] = EventStream.create();
    return stream;
}
function merged(array) {
    return new EventStream((next) => {
        const listeners = array.map((stream) => stream.listen(next));
        return () => listeners.forEach((x) => x());
    });
}

function held(stream, initial) {
    const [behavior, setBehavior] = mutValue(initial);
    mount(() => stream.listen(setBehavior));
    return behavior;
}
function reduced(stream, reducer, initial) {
    const [behavior, setBehavior] = mutValue(initial);
    mount(() => stream.listen((next) => {
        setBehavior(reducer(behavior.value, next));
    }));
    return behavior;
}

function read(runnable) {
    const data = createData(0);
    accessData(data); // log read.
    try {
        let sync = true;
        const disposer = untrack(() => executeRunnable(runnable, (value) => {
            if (sync) {
                throw {
                    $$RESOLVE: true,
                    $$VALUE: value,
                };
            }
            // notify change!
            setData(data, 0, false);
        }, (error) => {
            if (sync) {
                throw {
                    $$REJECT: true,
                    $$VALUE: error,
                };
            }
            // notify change!
            setData(data, 0, false);
        }));
        sync = false;
        throw new Suspend(disposer);
    }
    catch (e) {
        if (e.$$RESOLVE === true) {
            return e.$$VALUE;
        }
        else if (e.$$REJECT === true) {
            throw e.$$VALUE;
        }
        else {
            throw e; // not likely to happen.
        }
    }
}
function tryRead(runnable, fallback, onError) {
    try {
        return read(runnable);
    }
    catch (e) {
        if (e instanceof Suspend) {
            throw new SuspendWithFallback(fallback, e.cancel);
        }
        else {
            throw new SuspendWithFallback(onError(e), noop);
        }
    }
}
function resource() {
}

function debounce(time) {
    let timeoutId = undefined; // number or NodeJS.Timeout
    return (next) => {
        return (payload) => {
            if (timeoutId !== undefined) {
                clearTimeout(timeoutId);
            }
            timeoutId = setTimeout(() => {
                next(payload);
                timeoutId = undefined;
            }, time);
        };
    };
}
function throttle(time) {
    let timeoutId = undefined;
    let lastFired = 0;
    return (next) => {
        return (payload) => {
            if (Date.now() - lastFired > time) {
                next(payload);
                lastFired = Date.now();
            }
            else {
                if (timeoutId !== undefined) {
                    clearTimeout(timeoutId);
                }
                timeoutId = setTimeout(() => {
                    next(payload);
                    lastFired = Date.now();
                    timeoutId = undefined;
                }, Date.now() - lastFired);
            }
        };
    };
}
const animation = (next) => {
    let rafId = 0;
    return (payload) => {
        // last rafId is auto disposed
        rafId = requestAnimationFrame(() => {
            next(payload);
            rafId = 0;
        });
        return () => cancelAnimationFrame(rafId);
    };
};
function scheduled(source, scheduler) {
    const [dest, emitDest] = stream();
    mount(() => source.listen(scheduler(emitDest)));
    return dest;
}

class Reference {
    constructor(_current = null, _setter) {
        this._current = _current;
        this._setter = _setter;
    }
    get bind() {
        return (x) => {
            this._current = this._setter(x);
        };
    }
    set bind(x) {
        this._current = this._setter(x);
    }
    get current() {
        return this._current;
    }
    set current(value) {
        this._current = this._setter(value);
    }
}
function reference(initialValue, setter = identity) {
    return new Reference(initialValue, setter);
}

function testBed(setup) {
    const testScope = new Scope();
    const endScope = testScope.beginScope();
    const binding = setup(interactFn);
    endScope();
    return binding;
}
function interactFn(interactLogic) {
    return {
        expectEffects(runnable) {
            return __awaiter(this, void 0, void 0, function* () {
                const detach = Scope.current.attach();
                const expect = start(runnable()); // do side effects first
                start(interactLogic()); // start interact
                yield expect;
                detach();
            });
        },
    };
}

function isCell(value) {
    return value instanceof Cell;
}
function isEventStream(value) {
    return value instanceof EventStream;
}
function action(fn) {
    return (...args) => transaction(() => fn(...args));
}
function watch(cell, sideEffect, options) {
    mount(() => cell.watch(sideEffect, options));
}
function listen(stream, handler) {
    mount(() => stream.listen(handler));
}
function autorun(sideEffect) {
    mount(() => {
        const l = lazy();
        return l.watch(() => {
            return l.execute(sideEffect);
        }, { immediate: true });
    });
}

export { CanceledError, CancellablePromise, Cell, ComputationalCell, EventStream, Lazy, Reference, Scope, Semaphore, Suspend, SuspendWithFallback, Symbol_observable, Token, __current_collecting, action, all, allSettled, animation, any, autorun, combined, computed, constant, debounce, delay, mount as effect, executeRunnable, held, inject, isCell, isEventStream, lazy, listen, merged, mount, mutable as mut, mutArray, mutMap, mutSet, mutValue, mutable, never, nextAnimationFrame, provide, race, read, readEvents, reduced, reference, resolve, resource, scheduled, start, stream, suspended, task, testBed, throttle, timeout, transaction, tryRead, untrack, watch };
